/***********************************************************/
/* WCPsimul: a WWW Cache Removal Policy Simulation tool    */
/* Copyright (C) 1998 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/
// Simulation de cache WWW: calcul du request hitrate,  byte hitrate et latency ratio
// Nicolas Niclausse 
#define VERSION "Version 0.30 (last modification: <17/08/1999 17:13 Nicolas.Niclausse@sophia.inria.fr>)"

#include <iostream.h>
#include <String.h>
#include <fstream.h>
#include <std.h>
#include <GetOpt.h>
#include "WCPsimul.h"
#include "Cache.h"
#include <sys/times.h>
#include <math.h>
#include <limits.h>

// display usage information 
void usage(char *progname,int i) {
  cout << "Usage: " << progname <<" <arg> [-h] [-f <file>] [-i <file>] [-b <int>] [-o <int>] [-p <char>] [-a <int>] [-m <real>] [-l <real>] [-w <real>]  [-d] [-s <int>] [-c]\n";
  cout << VERSION << "\n";
  cout << " -h         give this help\n";
  cout << " -f <file>  input file\n";
  cout << " -i <file>  initialisation file (squid native format) (optional)\n";
  //  cout << " -s <int>   initialisation file use static (persistent) documents \n";
  //  cout << "                 <int> = size on the static cache\n";
  cout << " -b <int>   size of the cache (in kbytes)\n";
  cout << " -o <int>   maximum object size (in kbytes; default: 15MB)\n";
  cout << " -p <char>  politic of the garbage collector\n";
  cout << "                 r = LRU (default)\n";
  cout << "                 u = MIX\n";
  cout << "                 f = LFU\n";
  cout << "                 b = size\n";
  cout << "                 v = HYB\n";
  cout << "                 g = Greedy(size)\n";
  cout << "                 k = Greedy(MIX)\n";
  //  cout << "                 j = Bolot/Hoschka\n";
  //  cout << "                 a = LFU*_Aging\n";
  //  cout << " -t         don't cache one-timer requests\n";
  //  cout << " -a <int>   value of Amax for  LFU*_Aging policy (default: 10)\n";
  cout << " -v <int>   value of CONN for HYB policy (default: 2048)\n";
  cout << " -x <real>  value of BETA for MIX policy (default: 1.0)\n";
  cout << " -y <real>  value of  ALPHA for MIX policy (default: 1.0)\n";
  cout << " -z <real>  value of GAMMA for MIX policy (default: 1.0)\n";
  cout << " -m <real>  cache_size_high (default: 0.95)\n";
  cout << " -l <real>  cache_size_low (default: 0.90)\n";
  cout << " -w <real>  bw_hit (default: 100kB/s)\n";
  cout << " -c         disable cache_high/low mechanism\n";
  cout << " -e         start sampling hitrate after filling the cache\n";
  cout << "               (default: start sampling immediately)\n";
  cout << " -d         debug\n";
  cout << " -O <file>  output state of the cache at the end of the simulation\n";
  cout << " Output format: cachesize hitrate meanlatency bytehitrate meanlatency_bw cache_size_low cache_size_high CPU_elapsed MIX_beta MIX_alpha MIX_gamma latency_ratio cachesize_relative %CGI\n";
  exit(i);
}

// ----------------------------------------------------------------------------
//                                 Main                                      
// ----------------------------------------------------------------------------

int main(int argc, char* argv[])
{
  char * filename; // fichier logs
  char * outputname; // output cache
  char * fileinit; // fichier d'initialisation du cache
  int output = 0; // don't output cache by default
  Log_URL tmp;
  Log_URL * indice;
  long gccount = 0; // nombre d'appels de la procedure de garbage collection
  double mem_libre = 0; // memoire libre
  double tot_size = 0; // taille du cache en ??
  long logs = 0;
  double obj_max_size = 0; // taille maximum d'un objet dans le cache
  String IP,info,code; 
  double date;
  char politique = 'r';
  double limit_high = 0.95; // high water mark
  double limit_low = 0.90; // low water mark
  double debit_hit = 100;
  GetOpt getopt(argc,argv,"b:m:l:p:v:i:o:a:w:s:x:y:z:O:f:hdecCt");
  int option_char;
  int init = 0; // init option
  double tot_size_init = 0; 
  double logs_init = 0;
  int debug = 0; // debug option
  int high_low = 0; // option : disable cache_high/low mechanism
  int start = 1; // 0 si on veut commencer stats apres que le cache soit remplit
  // ie apres le premier appel a gc()
  // 1 sinon
  
  clock_t times_return; // variable pour mesurer temps d'execution
  struct tms buffer; // variable pour mesurer temps d'execution

  long cgi = 0; // requetes CGI
  long ims = 0; // requetes 304

  long conn = 2048; // politique HYB
  double w_n = 0.9 ; // politique HYB 
  long w_b = 8196; // politique HYB
  double beta = 1.0; // politique MIX
  double alpha = 1.0; // politique MIX
  double gamma = 1.0; // politique MIX
  //  double actual_time;
  Cache * cache; // Objet Cache

   cerr.precision(16);
  if (argc < 2)
   exit(1);

 while ((option_char=getopt()) != EOF)
   {
     switch (option_char)
       {
       case 'f':
	 filename=getopt.optarg;
	 break;
       case 'O':
	 outputname=getopt.optarg;
	 output=1;
	 break;
       case 'p':
	 politique=*getopt.optarg;
	 break;
       case 'x':
	 beta = atof(getopt.optarg);
	 break;
       case 'y':
	 alpha = atof(getopt.optarg);
	 break;
       case 'z':
	 gamma = atof(getopt.optarg);
	 break;
       case 'b':
	 mem_libre=1024*atof(getopt.optarg);
	 break;
       case 'o':
	 obj_max_size=1024*atof(getopt.optarg);
	 break;
       case 'a':
	 //	 amax=atoi(getopt.optarg);
	 break;
       case 'v':
	 conn=atoi(getopt.optarg);
	 break;
       case 'm':
	 limit_high=atof(getopt.optarg);
	 break;
       case 'l':
	 limit_low=atof(getopt.optarg);
	 break;
       case 'w':
	 debit_hit=atof(getopt.optarg);
	 break;
       case 'h':
	 usage(argv[0],0);
	 break;
       case 'e':
	 start = 0;
	 break;
       case 't':
	 //	 one_timer=1;
	 break;
       case 's':
	 //	 tot_size_static=1024*atof(getopt.optarg);
	 //	 persistent='y';
	 break;
       case 'c':
	 high_low=1;
	 break;
       case 'd':
	 debug=1;
	 break;
       case 'i':
	 fileinit=getopt.optarg;
	 init=1;
	 break;
       default:
	 usage(argv[0],1);
	 break;
       }
   }   

 ifstream in(filename);

 if (!in)
   {
     cerr << "erreur: fichier non trouve " << "\n";
     exit(1);
   }
 if (mem_libre <= 0)
   {
     cout << "0 1 0 1 0\n";
     exit(1);
   }

 gccount = 0;
 if (obj_max_size == 0)
   obj_max_size = min(OBJECT_MAXIMUM_SIZE,mem_libre/2);
 else
   obj_max_size = min(obj_max_size,mem_libre);
   
 cout.precision(12);

 switch (politique)
   {
   case 'v':
     cache = new Cache_aw(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, conn, w_b, w_n, start);
     break;
   case 'f':
     cache = new Cache_freq(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     break;
   case 'b':
     //     cache = new Cache_dyn(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     cache = new Cache_size(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     break;
   case 'u':
     cache = new Cache_mix(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start, beta, alpha, gamma);
     break;
   case 'g':
     cache = new Cache_greedy(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     break;
   case 'k':
     cache = new Cache_greedy_mix(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     break;
   default:
     cache = new Cache(mem_libre, limit_low, limit_high, obj_max_size, high_low, debit_hit, start);
     break;
   }
 if (debug) cache->debug_on();

// ----------------------------------------------------------------------------
 if (init) // initialisation du cache
   {
     ifstream init_file(fileinit);
     while (init_file >> tmp.date_last_ref >> tmp.retrieve_time >> IP >> info >> tmp.size >> code >> tmp.URL)
       {
	 if (tmp.size < cache->object_max_size())
	   if (tmp.size < cache->memoire_libre())
	     {
	       if (IP == "SIMU") //init from previous simulation
		 tmp.frequency = atoi(info);
	       indice = &tmp;
	       if (!(cache->search(indice)))
		 {
		   cache->insert_init(tmp);
		   tot_size_init += tmp.size;
		   if (debug) tmp.print();
		   if (debug) cerr << tot_size_init <<"\n";
		   logs_init++;
		 }
	     }
       }
   }
 
 // if (debug) cache->print();
   
// ----------------------------------------------------------------------------
// -----
// -----
// ----------------------------------------------------------------------------

 while (in >> date >> tmp.retrieve_time >> IP >> info >> tmp.size >> code >> tmp.URL )
   {
     tmp.date_last_ref = date*1000; // date du debut de requete en ms
     //     tmp.persistent='n';
     //     tmp.politique=politique;
     if (tmp.retrieve_time > 0)
       {
	 logs++;

	 if ((tmp.URL.contains('?'))
	     || (tmp.URL.contains("cgi-bin"))) //URL contient ? donc script CGI
	   {
	     tot_size += tmp.size;
	     cache->overflow_miss(tmp); //hum, il faudrait une procedure particuliere pour cgi
	     cgi++;
	     if (debug) cerr << "# CGI: " << tmp.URL << "\n";
	   }
	 else
	   {
	     //if (debug) tmp.print();
	     indice = &tmp;
	     if (cache->search(indice))
	       //hit ou document modifie
	       {
		 if (!(info.contains("200"))) //requete 304 ou assimile
		     {
		       ims++;
		       tmp.size = indice->size;
		       tmp.retrieve_time = indice->retrieve_time;
		     };
		 if (indice->size != tmp.size) // taille differente
		   {
		     cache->insert_newsize(tmp,indice);
		     if (cache->active_gc())
		       {
			 if (debug)
			   cerr << "# mem_libre <"<< 100*cache->HIT_high() << "% mem totale \n";
			 cache->gc(tmp.date_last_ref+tmp.retrieve_time, tmp);
		       }
		   }
		 else //hit
		   cache->insert_hit(tmp,indice);
	       }
	     else //miss
	       {
		 if (tmp.size > cache->object_max_size()) // document trop grand
		   cache->overflow_miss(tmp);
		 else
		   {
		     if (tmp.size > cache->memoire_libre())
		       cache->gc(tmp.date_last_ref+tmp.retrieve_time, tmp);
		     cache->insert_miss(tmp);
		     if (cache->active_gc())
		       {
			 if (debug)
			   cerr << "# mem_libre <"<< 100*cache->HIT_high() << "% mem totale \n";
			 cache->gc(tmp.date_last_ref+tmp.retrieve_time,tmp);
		       }
		   }
	       }
	     tot_size += tmp.size;
	   }
       }
   }
// ----------------------------------------------------------------------------

 if (output)
   {
     ofstream output_file(outputname);
     cache->output(output_file);
   }

 // cout << "# number of requests: " << logs << " "<<cache->first_time_tot()<<"\n";
 cout << "# number of requests: " << logs <<"\n"
      << "# number of sampling requests: " << cache->sampling() <<"\n"
      << "# number of GC calls: " << cache->gccount << "\n"
      << "# number of cache overflow: " << cache->overflow() << "\n"
      << "# number of ims: " << ims << "\n"
      << "# object_max_size: " << obj_max_size << "\n"
      << "# Total bytes transfered: "<< tot_size << "\n"
      << "# Newssize rate: " << cache->newsizerate() << "\n"
      << "# init_size: " << tot_size_init << " init_files: " << logs_init << "\n"
      << "# politique: " << politique;
 // if (politique == 'a')
 //   cout << "  Amax=" << amax <<"\n";
 // else
 cout << "\n";

 cout << mem_libre/1024 << " "; 		// capacit du cache
 cout.precision(5);
 cout << 1-cache->missrate() << " " 		// hitrate
      << cache->w_missrate() << " " 		// mean latency
      << 1-cache->ws_missrate()/double(tot_size)// byte hitrate
      << " " << cache->w_hitrate() + cache->w_missrate(); //
 if (high_low == 1)
   cout << " " << 1 << " " << 1;
 else
   cout << " " << limit_low	//seuil min
	<< " " << limit_high; 	//seuil max
 
 times_return = times(&buffer);
 cout << " " << double(buffer.tms_utime+buffer.tms_stime)/CLK_TICKS	// temps CPU
 // cout << " " << long(tot_size_static/1024.0) << "\n";
      << " " << beta			// parametre beta politique MIX
      << " " << alpha			// parametre alpha politique MIX
      << " " << gamma			// parametre gamma politique MIX
      << " " << cache->win_time()	// 
      << " " << double(mem_libre)/double(tot_size) //
      << " " << double(cgi)/logs		// scripts CGI
      << "\n";
}
