// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef __GNUG__
#pragma implementation
#endif
#include "String.Couple.CHMap.h"
#include <String.h>

//extern unsigned int hash(String&);
extern unsigned int hash(String& s)
{
  long j = s.length();
  int i,n;
  
  for (i = j / 2, n = 0; i < j; i++)
    n ^= 271 * (unsigned) s[i];
  i = n ^ (j * 271);
  i=(i < 0 ? -i : i);
  return i ;
  //   return (i % DEFAULT_INITIAL_CAPACITY);
  // **************** WARNING !!! ****************
  // j'ai enleve le modulo -> ??????? 
}


// The nodes are linked together serially via a version
// of a trick used in some vtables: odd pointers are
// actually links to the next table entry. 
// Not terrible, but not wonderful either

static inline int goodCHptr(StringCoupleCHNode* t)
{
  return ((((unsigned)t) & 1) == 0);
}

static inline StringCoupleCHNode* index_to_CHptr(int i)
{
  return (StringCoupleCHNode*)((i << 1) + 1);
}

static inline int CHptr_to_index(StringCoupleCHNode* t)
{
  return ( ((unsigned) t) >> 1);
}

StringCoupleCHMap::StringCoupleCHMap(Couple& dflt, unsigned int sz)
     :StringCoupleMap(dflt)
{
  tab = (StringCoupleCHNode**)(new StringCoupleCHNodePtr[size = sz]);
  for (unsigned int i = 0; i < size; ++i) tab[i] = index_to_CHptr(i+1);
  count = 0;
}

StringCoupleCHMap::StringCoupleCHMap(StringCoupleCHMap& a) :StringCoupleMap(a.def)
{
  tab = (StringCoupleCHNode**)(new StringCoupleCHNodePtr[size = a.size]);
  for (unsigned int i = 0; i < size; ++i) tab[i] = index_to_CHptr(i+1);
  count = 0;
  for (Pix p = a.first(); p; a.next(p)) (*this)[a.key(p)] = a.contents(p);
}


Pix StringCoupleCHMap::seek(String& key)
{
  unsigned int h = StringHASH(key) % size;

  for (StringCoupleCHNode* t = tab[h]; goodCHptr(t); t = t->tl)
    if (StringEQ(key, t->hd))
      return Pix(t);

  return 0;
}


Couple& StringCoupleCHMap::operator [](String& item)
{
  unsigned int h = StringHASH(item) % size;
  StringCoupleCHNode* t;

  for (t = tab[h]; goodCHptr(t); t = t->tl)
    if (StringEQ(item, t->hd))
      return t->cont;

  t = new StringCoupleCHNode(item, def, tab[h]);
  tab[h] = t;
  ++count;
  return t->cont;
}


void StringCoupleCHMap::del(String& key)
{
  unsigned int h = StringHASH(key) % size;

  StringCoupleCHNode* t = tab[h]; 
  StringCoupleCHNode* trail = t;
  while (goodCHptr(t))
  {
    if (StringEQ(key, t->hd))
    {
      if (trail == t)
        tab[h] = t->tl;
      else
        trail->tl = t->tl;
      delete t;
      --count;
      return;
    }
    trail = t;
    t = t->tl;
  }
}


void StringCoupleCHMap::clear()
{
  for (unsigned int i = 0; i < size; ++i)
  {
    StringCoupleCHNode* p = tab[i];
    tab[i] = index_to_CHptr(i+1);
    while (goodCHptr(p))
    {
      StringCoupleCHNode* nxt = p->tl;
      delete(p);
      p = nxt;
    }
  }
  count = 0;
}

Pix StringCoupleCHMap::first()
{
  for (unsigned int i = 0; i < size; ++i) if (goodCHptr(tab[i])) return Pix(tab[i]);
  return 0;
}

void StringCoupleCHMap::next(Pix& p)
{
  StringCoupleCHNode* t = ((StringCoupleCHNode*)p)->tl;
  if (goodCHptr(t))
    p = Pix(t);
  else
  {
    for (unsigned int i = CHptr_to_index(t); i < size; ++i) 
    {
      if (goodCHptr(tab[i]))
      {
        p =  Pix(tab[i]);
        return;
      }
    }
    p = 0;
  }
}


int StringCoupleCHMap::OK()
{
  int v = tab != 0;
  int n = 0;
  for (unsigned int i = 0; i < size; ++i)
  {
    StringCoupleCHNode* p;
    for (p = tab[i]; goodCHptr(p); p = p->tl) ++n;
    v &= CHptr_to_index(p) == i + 1;
  }
  v &= count == n;
  if (!v) error("invariant failure");
  return v;
}
