/***********************************************************/
/* WCPsimul: a Web Cache removal Policy simulation tool    */
/* Copyright (C) 1996 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/
//
// Classe de Tableaux
// 
// Nicolas Niclausse
// Last update: 18/06/96
// Date de cration: 16/10/95

#include "TTableau.h"
#include <iostream.h>

template<class T> TTableau<T>::TTableau()
{
  size=0;
}

template<class T> TTableau<T>::TTableau( long i)
{
  size=i;
  if (size!=0)
    tab=new T[size];
}

template<class T> void TTableau<T>::apply_fct(T f(T))
{
  for ( long i=0;i<size;i++)
    tab[i]=f(tab[i]);
}

template<class T> TTableau<T>::~TTableau()
{
  if (size!=0)
    delete[] tab;
}

template<class T>  long TTableau<T>::maxsize() const
{
  return size;
}

template<class T> void TTableau<T>::initialise(const T s)
{
  long i;
  
  if (size<=0)
    {
      if (size!=0 )
	{ 
	  cerr << "Erreur: indice negatif !!!\n" ;
	  exit(1);
	}
    }
  else
    {
      for (i=0;i<size;i++)
	tab[i]=s;
    }
}

template<class T> T& TTableau<T>::operator [](long n) const
{
  if (n<size)
    if (n<0)
      {
	cerr << "Erreur(TTableau.operator[]): indice negatif !!!" << n <<"\n" ; 
	exit(1);
      }
    else
      if (size!=0)
	return tab[n];
      else
	{
	  cerr << "Erreur(TTableau.operator[]): vecteur vide\n" ;
	  exit(1);
	}
  else
    {
      cerr << "Erreur: indice superieur a la dimension du vecteur \n" ;
      cerr << "        valeur de  l'indice: " << n << "\n" ;
      cerr << " Dimension du Vecteur: " << size << "\n" ;
      exit(1);
    }
}

template<class T> void TTableau<T>::resize(long n)
{
  long i,j;
  T * tmp;

  if (n==0)
    {
      if (size!=0)
	delete[] tab;
      size=0;
    }
  else if (n<0)
    {
      cerr << " Erreur(TTableau.resize(long)): dimension negative !!!: " << n << "\n" ;
      exit(1);
    }
  else if (size==0)
    {
      size=n;
      tab =new T[size];
    }
  
  else if (n!=size)
    {
      tmp =new T[size];
      for (i=0;i<size;i++)
	tmp[i]=tab[i];
      if (n>size)
	j=size;
      else
	j=n;
      size=n;
      delete[] tab;
      tab =new T[size];
      
      for (i=0;i<j;i++)
	tab[i]=tmp[i];
      delete[] tmp;
    }
};

template<class T> void TTableau<T>::resize(long n, long m)
{
  //Enleve element compris entre n et m (inclus)
  
  long i,taille=m-n;
  T * tmp;

  if (taille<0)
    {
      cerr <<"n>m dans resize(n,m) erreur\n";
      exit(1);
    }
  else  if (n<0)
    {
      cerr <<"n<0 dans resize(n,m) erreur\n";
      exit(1);
    }
  else if (m>=size)
    {
      cerr <<"m>size dans resize(n,m) erreur\n";
      exit(1);
    }
  else if ((size-taille-1)==0)
    {
      size=0;
      delete[] tab;
    }
  else
    {
      tmp=new T[size];
      for (i=0;i<size;i++)
	tmp[i]=tab[i];
      delete[] tab;
      tab =new T[size-taille-1];
      if (n>0)
	for (i=0;i<n;i++)
	  tab[i]=tmp[i];
      if (m<size-1)
	for (i=m+1;i<size;i++)
	  tab[i-m+n-1]=tmp[i];
      delete[] tmp;
      size-=taille+1;
    }
}

template<class T> void TTableau<T>::swap(long i, long j, long n)
{
  T tmp;
  long k;
  
  if (size!=0)
    {
      if ((i+n<=j)&&(j+n<=size))
	{
	  for (k=0;k<n;k++)
	    {
	      tmp=tab[i+k];
	      tab[i+k]=tab[j+k];
	      tab[j+k]=tmp;
	    }
	}
      else
	{
	  cerr << "swap:erreur d'indice\n";
	  cerr << i << " " <<  j << " "<< n << "\n";
	  exit(1);
	}
    }
}

template<class T> void TTableau<T>::operator = (const TTableau<T>& m)
{

  long taille=m.maxsize();

  if (size!=taille)
    {
      if (taille!=0)
	{
	  if (size!=0)
	    delete[] tab;
	  tab = new T[taille];
	}
      size=taille;
    }
  if (size!=0)
    for (long i=0;i<size;i++)
      tab[i]=m[i];
};

template<class T> void TTableau<T>::decalage(const long n)
{
  // decalage du vecteur de n indices (n peut etre negatif)
  // diminue la taille du vecteur en sortie
  // ex: [a,b,c,d] avec param 2 --> [a,b]
  //si n<0 cela revient a couper le vecteur
  long  tmp=size-n;
  long  i;

  if (size!=0)
    {
      if (tmp!=0)
	{
	  if (tmp>0)
	    {
	      if (n>0)
		for (i=0;i<tmp;i++)
		  tab[i]=tab[i+n];
	      this->resize(tmp);
	    }
	  else
	    {
	      this->resize(-tmp);
	    }
	}
      else
	{
	  size=0;
	  delete[] tab;
	}
    }
}

template<class T> void TTableau<T>::insert(const TTableau<T>& x,  long k)
{
  //insere le vecteur x a partir de l'indice j
  // ex: inserer [a,b] dans [i,j,k,l] a la position 2 donne
  // [i,j,a,b,k,l]
  
   long n=x.maxsize();
   long m=size;
   long i;

  if (k>size)
    {
      cerr << "In insert(x,k): k>size "<< k << " "<< size << "\n";
      exit(1);
    }
  else
    {
      this->resize(m+n);
      if ((m!=0)&&(k!=m))
	{
	  for (i=m-1;i>k;i--)
	    tab[i+n]=tab[i];
	  tab[k+n]=tab[k];
	  cout << "In insert(x,k): k="<< k << " n="<< n << " m="<< m<< "\n";
	}
      for (i=0;i<n;i++)
	tab[i+k]=x[i];
    }
}


template<class T> void TTableau<T>::insert(const TTableau<T>& x,  long a, long b)
{
  //insere les b premiers elements de x a partir de l'indice a
  // ex: inserer 2 el de [a,b,c] dans [i,j,k,l] a la position 2 donne
  // [i,j,a,b,k,l]
  long i;
  
  this->resize(size+b);
  for (i=a;i<size;i++)
    {
      tab[i+b]=tab[i];
    }
  for (i=0;i<b;i++)
    {
      tab[i+a]=x[i];
    }
}

//---------------------------------------------------------------------------

template<class T> TTableau<T>::TTableau(const TTableau<T>& m)
{

  size=m.maxsize();
  if (size!=0)
    {
      tab=new T[size];
      for (long i=0;i<size;i++)
	tab[i]=m.tab[i];
    }
}

template<class T> static int gsort (T *base_ptr, int total_elems, int cmp(T&,T&));

template<class T> void TTableau<T>::sort(int compar(T&,T&), long len)
 {
   gsort(tab, len, compar);
};

 
template<class T> static inline void SWAP(T* A, T* B)
{
  T tmp = *A; *A = *B; *B = tmp;
};


//An adaptation of Schmidt's new quicksort


/* This should be replaced by a standard ANSI macro. */
#define BYTES_PER_WORD 8
#define BYTES_PER_LONG 4

/* The next 4 #defines implement a very fast in-line stack abstraction. */

#define STACK_SIZE (BYTES_PER_WORD * BYTES_PER_LONG)
#define PUSH(LOW,HIGH) do {top->lo = LOW;top++->hi = HIGH;} while (0)
#define POP(LOW,HIGH)  do {LOW = (--top)->lo;HIGH = top->hi;} while (0)
#define STACK_NOT_EMPTY (stack < top)                

/* Discontinue quicksort algorithm when partition gets below this size.
   This particular magic number was chosen to work best on a Sun 4/260. */
#define MAX_THRESH 4


/* Order size using quicksort.  This implementation incorporates
   four optimizations discussed in Sedgewick:
   
   1. Non-recursive, using an explicit stack of pointer that
      store the next array partition to sort.  To save time, this
      maximum amount of space required to store an array of
      MAX_INT is allocated on the stack.  Assuming a 32-bit integer,
      this needs only 32 * sizeof (stack_node) == 136 bits.  Pretty
      cheap, actually.

   2. Chose the pivot element using a median-of-three decision tree.
      This reduces the probability of selecting a bad pivot value and 
      eliminates certain extraneous comparisons.

   3. Only quicksorts TOTAL_ELEMS / MAX_THRESH partitions, leaving
      insertion sort to order the MAX_THRESH items within each partition.  
      This is a big win, since insertion sort is faster for small, mostly
      sorted array segements.
   
   4. The larger of the two sub-partitions is always pushed onto the
      stack first, with the algorithm then concentrating on the
      smaller partition.  This *guarantees* no more than log (n)
      stack size is needed! */


template<class T> struct stack_node {  T *lo;  T *hi; };

template<class T> static int gsort (T *base_ptr, int total_elems, int cmp(T&,T&))
{
/* Stack node declarations used to store unfulfilled partition obligations. */
  T   pivot_buffer;
  int   max_thresh   = MAX_THRESH;
  //  struct stack_node {  T *lo;  T *hi; };


  if (total_elems > MAX_THRESH)
    {
      T       *lo = base_ptr;
      T       *hi = lo + (total_elems - 1);
      T       *left_ptr;
      T       *right_ptr;
      stack_node<T> stack[STACK_SIZE]; /* Largest size needed for 32-bit int!!! */
      //      stack_node stack; /* Largest size needed for 32-bit int!!! */
      stack_node<T> *top = stack;

      for (;;)
        {
          {
            T *pivot = &pivot_buffer;
            {
              /* Select median value from among LO, MID, and HI. Rearrange
                 LO and HI so the three values are sorted. This lowers the 
                 probability of picking a pathological pivot value and 
                 skips a comparison for both the LEFT_PTR and RIGHT_PTR. */

              T *mid = lo + ((hi - lo) >> 1);

              if ((*cmp) (*mid, *lo) < 0)
                SWAP (mid, lo);
              if ((*cmp) (*hi, *mid) < 0)
              {
                SWAP (mid, hi);
                if ((*cmp) (*mid, *lo) < 0)
                  SWAP (mid, lo);
              }
              *pivot = *mid;
              pivot = &pivot_buffer;
            }
            left_ptr  = lo + 1;
            right_ptr = hi - 1; 

            /* Here's the famous ``collapse the walls'' section of quicksort.  
               Gotta like those tight inner loops!  They are the main reason 
               that this algorithm runs much faster than others.

	       [Yeah, but there is a function call in the inner loop,
	       which makes the speed claim rather dubious.  --Per Bothner] */
            do 
              {
                while ((*cmp) (*left_ptr, *pivot) < 0)
                  left_ptr += 1;

                while ((*cmp) (*pivot, *right_ptr) < 0)
                  right_ptr -= 1;

                if (left_ptr < right_ptr) 
                  {
                    SWAP (left_ptr, right_ptr);
                    left_ptr += 1;
                    right_ptr -= 1;
                  }
                else if (left_ptr == right_ptr) 
                  {
                    left_ptr += 1;
                    right_ptr -= 1;
                    break;
                  }
              } 
            while (left_ptr <= right_ptr);

          }

          /* Set up pointers for next iteration.  First determine whether
             left and right partitions are below the threshold size. If so, 
             ignore one or both.  Otherwise, push the larger partition's
             bounds on the stack and continue sorting the smaller one. */

          if ((right_ptr - lo) <= max_thresh)
            {
              if ((hi - left_ptr) <= max_thresh)
		{ /* Ignore both small partitions. */
		  if (STACK_NOT_EMPTY)
		    POP (lo, hi);
		  else
		    break;  /* No pending partitions - we're done. */
		}
              else              /* Ignore small left partition. */  
                lo = left_ptr;
            }
          else if ((hi - left_ptr) <= max_thresh) /* Ignore small right partition. */
            hi = right_ptr;
          else if ((right_ptr - lo) > (hi - left_ptr)) /* Push larger left partition indices. */
            {                   
              PUSH (lo, right_ptr);
              lo = left_ptr;
            }
          else                  /* Push larger right partition indices. */
            {                   
              PUSH (left_ptr, hi);
              hi = right_ptr;
            }
        }
    }

  /* Once the BASE_PTR array is partially sorted by quicksort the rest
     is completely sorted using insertion sort, since this is efficient 
     for partitions below MAX_THRESH size. BASE_PTR points to the beginning 
     of the array to sort, and END_PTR points at the very last element in
     the array (*not* one beyond it!). */


  {
    T *end_ptr = base_ptr + 1 * (total_elems - 1);
    T *run_ptr;
    T *tmp_ptr = base_ptr;
    T *thresh  = (end_ptr < (base_ptr + max_thresh))? 
      end_ptr : (base_ptr + max_thresh);

    /* Find smallest element in first threshold and place it at the
       array's beginning.  This is the smallest array element,
       and the operation speeds up insertion sort's inner loop. */

    for (run_ptr = tmp_ptr + 1; run_ptr <= thresh; run_ptr += 1)
      if ((*cmp) (*run_ptr, *tmp_ptr) < 0)
        tmp_ptr = run_ptr;

    if (tmp_ptr != base_ptr)
      SWAP (tmp_ptr, base_ptr);

    /* Insertion sort, running from left-hand-side up to `right-hand-side.' 
       Pretty much straight out of the original GNU qsort routine. */

    for (run_ptr = base_ptr + 1; (tmp_ptr = run_ptr += 1) <= end_ptr; )
      {

        while ((*cmp) (*run_ptr, *(tmp_ptr -= 1)) < 0)
          ;

        if ((tmp_ptr += 1) != run_ptr)
          {
            T *trav;

            for (trav = run_ptr + 1; --trav >= run_ptr;)
              {
                T c = *trav;
                T *hi, *lo;

                for (hi = lo = trav; (lo -= 1) >= tmp_ptr; hi = lo)
                  *hi = *lo;
                *hi = c;
              }
          }

      }
  }
  return 1;
}


//---------------------------------------------------------------------------







