/***********************************************************/
/* WCPsimul: a Web Cache removal Policy simulation tool    */
/* Copyright (C) 1997 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/

// Classe de Vecteurs
// Nicolas Niclausse

#include "TVecteur.h"
#include <iostream.h>
#include <stdio.h>
#include <math.h>
#define MIN(a,b) (a>b? b : a)
#define MAX(a,b) (a<b? b : a)

template<class T> TVecteur<T> TVecteur<T>::operator / (const T& m) const 
{
  TVecteur<T> x(size);
   long i;
  
  if (m!=0)
    {
      for (i=0;i<size;i++)
	{
	  x[i]=tab[i]/m;
	}  
      return x;
    }
  else
    {
      cerr << "Erreur(operateur / Classe TVecteur): Division par zero \n";
      exit(1);
    }
}

template<class T> TVecteur<T> TVecteur<T>::operator * (const T& m) const
{
  TVecteur<T> x(size);
   long i;
  
  for (i=0;i<size;i++)
    {
      x[i]=tab[i]*m;
    }  
  return x;
}

template<class T> void TVecteur<T>::initialise_sum(const T& m)
{
  long i;

  if (size<=0)
    {
      if (size!=0)
	cerr << "Erreur: dimension negative !!! \n ";
    }
  else
    for (i=0;i<size;i++)
      {
	tab[i]=i+m;
      }  
}

template<class T> TVecteur<T> TVecteur<T>::operator + (const TVecteur<T>& m)
{
   long taille=m.maxsize();
  TVecteur<T> x(taille);
   long i;
  
  if (taille!=size)
    {
      cerr << "Erreur: addition de vecteurs de dimension differente \n ";
      exit(1);
    }
  else
    {
      for (i=0;i<taille;i++)
	{
	  x[i]=m[i]+tab[i];
	}  
      return x;
    }
}

template<class T> TVecteur<T> TVecteur<T>::operator - (const TVecteur<T>& m)
{
   long taille=m.maxsize();
  TVecteur<T> x(size);
   long i;
  
  if (taille!=size)
    {
      cerr << "Erreur: soustraction de vecteurs de dimension differente \n";
      exit(1);
    }
  else
    {
      for (i=0;i<taille;i++)
	{
	  x[i]=m[i]-tab[i];
	}  
      return x;
    }
}


template<class T> double TVecteur<T>::moyenne() const
{
    double esperance=0.0;
     long i;

    for (i=0;i<size;i++)
	esperance+=tab[i];
    esperance/=size;
    return esperance;
};

template<class T> double TVecteur<T>::moyenne(const  long k,const   long j) const
{
    double esperance=0.0;
     long i;

    if ((k<j)&&(j<size))
      {
	for (i=k;i<=j;i++)
	  esperance+=tab[i];
	esperance/=j-k+1;
	return esperance;
      }
    else
      {
	if (j<size)
	  cerr << "Erreur(fonction Moyenne(j,k)): indice superieur a la dimension du vecteur \n";
	else
	  cerr << "Erreur(fonction Moyenne(j,k)): j>=k \n";
	cerr << " valeur des  indices: " << j << " " << k << "\n" ;
	cerr << " Dimension du Vecteurs: " << size << "\n" ;
	exit(1);
      }
  
};

template<class T> double TVecteur<T>::variance() const
{
    double esperance=0.0,var=0.0,c=0.0;
     long i;

    for (i=0;i<size;i++)
	{
	    c=tab[i]-esperance;
	    esperance+=c/(i+1);
	    var+=c*(tab[i]-esperance);
	};
    var/=size;
    return var;
};

template<class T> double TVecteur<T>::variance(  long k, long j) const
{
    double esperance=0.0,var=0.0,c=0.0;
     long i;
    
    if ((k<j)&&(j<size))
      {
	for (i=k;i<=j;i++)
	  {
	    c=tab[i]-esperance;
	    esperance+=c/(i-k+1);
	    var+=c*(tab[i]-esperance);
	  }
	var/=(j-k+1);
	return var;
      }
    else
      {
	if (j<size)
	  cerr << "Erreur(fonction Moyenne(j,k)): indice superieur a la dimension du vecteur \n";
	else
	  cerr << "Erreur(fonction Moyenne(j,k)): j>=k \n";
	cerr << "Valeur des  indices: "<< j << " " <<k << "\n" ;
	cerr << " Dimension du Vecteurs: " << size << "\n" ;
	exit(1);
      }
};

template<class T> double TVecteur<T>::ecart_type() const
{
    return sqrt(variance());
}

template<class T> T TVecteur<T>::max() const
{
  T x=tab[0];
  
  for ( long i=1;i<size;i++)
    if (tab[i]>x)
      x=tab[i];
  return x;
}

template<class T> T TVecteur<T>::min() const
{
  T x=tab[0];
  
  for ( long i=1;i<size;i++)
    if (tab[i]<x)
      x=tab[i];
  return x;
};

template<class T> T TVecteur<T>::somme() const
{
  T s=0;
  
  for ( long i=0;i<size;i++) s+=tab[i];
  return s;
};

template<class T> T TVecteur<T>::somme( long j) const
{
  T s=0;
  
  if (j<size)
    {
      for ( long i=0;i<j;i++) s+=tab[i];
      return s;
    }
  else
    {
      cerr << "Erreur(somme(j)): j superieur a la dimension du Vecteur \n";
      cerr << "                  Valeur de  l'indice: " << j << "\n" ;
      cerr << "                  Dimension du vecteur: " << size << "\n" ;
      exit(1);
    }
};

template<class T> TVecteur<T> TVecteur<T>::tri_fusion()
/*-----------------------------------------------------------------------------
  An implementation of a "fusion-sort".
  Written by Franois Fleuret. Oct 94, revised Oct 97.

  Please report bugs, critics and comments to francois.fleuret@inria.fr
  Thanks to Linux, XFree86, GNU C++ and Emacs
-----------------------------------------------------------------------------*/
{
   long t, i, j, a, b;
  T *work2, *tmp;
  TVecteur work1(size);
  
  if(size == 1)
    {
      work1[0] = tab[0];
    }
  else
    {
      
      for(j=0; j<=size-2; j+=2)
	if(tab[j]<tab[j+1])
	  {
	    work1[j] = tab[j];
	    work1[j+1] = tab[j+1];
	  }
	else
	  {
	    work1[j] = tab[j+1];
	    work1[j+1] = tab[j];
	  };

      if(j<size) work1[j] = tab[j];

      work2 = new T[size];

      for(t=2; t<size; t*=2)
	{
	  tmp = work1.tab; work1.tab = work2; work2 = tmp;

	  for(i=0; i<=size-2*t; i+=2*t)
	    {
	      a = i; b = i+t; j = i;
	      if(work2[i+t-1]<work2[i+2*t-1])
		{
		  while(a<i+t)
		    if(work2[a]<work2[b]) work1[j++] = work2[a++];
		    else                  work1[j++] = work2[b++];
		  for(; j<i+2*t; j++) work1[j] = work2[b++];
		}
	      else
		{
		  while(b<i+2*t)
		    if(work2[a]<work2[b]) work1[j++] = work2[a++];
		    else                  work1[j++] = work2[b++];
		  for(; j<i+2*t; j++) work1[j] = work2[a++];
		};
	    };

	  if(size>i+t)
	    {
	      a = i; b = i+t; j = i;
	      if(work2[i+t-1]<work2[size-1])
		{
		  while(a<i+t)
		    if(work2[a]<work2[b]) work1[j++] = work2[a++];
		    else                  work1[j++] = work2[b++];
		  for(; j<size; j++) work1[j] = work2[b++];
		}
	      else
		{
		  while(b<size)
		    if(work2[a]<work2[b]) work1[j++] = work2[a++];
		    else                  work1[j++] = work2[b++];
		  for(; j<size; j++) work1[j] = work2[a++];
		};
	    }
	  else
	    for(j=i; j<size; j++) work1[j] = work2[j];

	};
      
      delete work2;
    };
  
  return work1;
};

// Calcul le graphique de variance
template<class T> int TVecteur<T>::graphique_variance(ostream& output, long mini, long maxi)
{
  const scale=2;
  const n=10;
  long m,size0,nmax,i=0,j,m0=1,taille=maxi-mini;
   long dep[n]={2,3,5,7,9,11,13,15,17,19};
  TVecteur<T> tab2(taille/2+1); //verifier quand taille tableau impaire ???
  
  nmax=long(log(taille/130.0)/log(2.0));
  output << 0.0 << " " << log10(this->variance(mini,maxi-1)) << "\n" ;
  for (j=0;j<n;j++)
    {
      tab2.initialise(0);
      size0=taille;
      for (i=0;i<size0;i++)
	tab2[i/dep[j]]+=tab[i+mini];
      size0/=dep[j];
      m0=dep[j];
      output << log10(m0) << " "<< log10(tab2.variance(0,size0-1)/(m0*m0)) <<"\n" ;
      for (m=2;m<=nmax;m++)
	{
	  size0/=scale;
	  for (i=0;i<size0;i++)
	    tab2[i]=tab2[i*scale]+tab2[i*scale+1];
	  m0*=scale;
	  output << log10(m0)<<" "<< log10(tab2.variance(0,size0-1)/(m0*m0)) <<"\n";
	}
    }
  return 0;
};


template<class T> TVecteur<T> TVecteur<T>::nombre_clients_GD1(T c, T init)
{
   long taille=size;
  TVecteur<T> w(taille+1);
   long i=0;
  
  w[0]=init;
  // w = Nombre de clients. File vide a l'instant initial t=0
  // arrivee(n) : nombre d'arrivee pendant l'intervalle (n-1,n)
  // c nombre de clients traite par unite de temps
  
  for(i=0;i<taille;i++)
    {
      w[i+1]=MAX(0,w[i]+tab[i]-c); // Lindley
    }
  return w;
};

template<class T> TVecteur<T> TVecteur<T>::departs_GD1(T c, T init)
{
  const taille=size;
  TVecteur<T> depart(taille),x(1);

  // arrivee(n) : nombre d'arrivee pendant l'intervalle (n-1,n)
  // c nombre de clients traite par unite de temps
  
  x=this->nombre_clients_GD1(c,init);
  for ( long i=0;i<taille;i++)
    depart[i]=MIN(c,x[i]+tab[i]);
  return depart;
}

template<class T> TVecteur<T> TVecteur<T>::nombre_clients_GD1B(T B,T c, T init)
{
   long taille=size;
  TVecteur<T> w(taille+1);
   long i=0;
  
  w[0]=init;
  // w = Nombre de clients. File vide a l'instant initial t=0
  // c nombre de clients traite par unite de temps
  // B taille du buffer
  
  for(i=0;i<taille;i++)
    {
      w[i+1]=MAX(0,MIN(B,w[i]+tab[i]-c)); // Lindley
    }
  return w;
};

template<class T> TVecteur<T> TVecteur<T>::departs_GD1B(T B,T c, T init)
{
  const taille=size;
  TVecteur<T> depart(taille),x(1);

  // c nombre de clients traite par unite de temps
  // B taille du buffer
  
  x=this->nombre_clients_GD1B(B,c,init);
  for ( long i=0;i<taille;i++)
    depart[i]=MIN(c,x[i]+tab[i]);
  return depart;
}

template<class T> double TVecteur<T>::lossrate_GD1B(T B,T c, T init)
{
  const taille=size;
  double loss=0.0;
  TVecteur<T> depart(taille),x(1);

  // c nombre de clients traite par unite de temps
  // B taille du buffer
  
  x=this->nombre_clients_GD1B(B,c,init);
  for ( long i=0;i<taille;i++)
    loss+=double(MAX(0,x[i]+tab[i]-c-x[i+1]));

  return loss/(this->moyenne()*taille);
}

template<class T> TVecteur<T> TVecteur<T>::temps_sejour(const TVecteur<T>& arrivee,const TVecteur<T>& service, T c)
{
  TVecteur<T> w(service.maxsize());

  w[0]=0; // w = temps d'attente. File vide a l'instant initial t=0
  
  for ( long i=0;i<arrivee.maxsize()-1;i++)
    {
      w[i+1]=MAX(0,w[i]+service[i]/c-arrivee[i+1]+arrivee[i]); // Lindley
    }

  w=w+service/c; //processus des temps de sejour
  return w;
}

template<class T> TVecteur<T> TVecteur<T>::departs_GG1(const TVecteur<T>& arrivee,const TVecteur<T>& service, T c)
{
  return temps_sejour(arrivee,service,c)+arrivee;
}

template<class T> TVecteur<T> TVecteur<T>::box2ps(const TVecteur<T>& x)
// Nombre par unit de temps ->echantillon processus stochastique (date d'arrive par ex)
{
  double s=0;
   long i,j;
  TVecteur<T> y(1);
  
  y.resize(long(x.somme())+1);
  for (i=0;i<x.maxsize();i++)
    {
      if (x[i]>0)
	for (j=1;j<=x[i];j++)
	  {
	    y[long(s)+j]=i;
	  }
      s+=x[i];
    }
  return y;
}

template<class T> TVecteur<T> TVecteur<T>::ps2box(const TVecteur<T>& x)
// echantillon processus stochastique (date d'arrive par ex)-> Nombre par unit de temps
{
   long i;
  T s;
  TVecteur<T> y(1);
  
  s=x.max();
  y.resize(long(s)+1);
  y.initialise(0);
  for (i=0;i<x.maxsize();i++)
    {
      y[long(x[i])]++; 
    }
  return y;
}
// Calcul du Rescaled Adjusted Range 
// !!! A modifier !!!

template<class T> int TVecteur<T>::RS(ostream& output)
{
    const m=10;
    const K=10;
    long n,n0,nmax,size_t,i=0,j,k,dep[m]={2,3,5,7,9,11,13,15,17,19};
    TVecteur<T> X(size),w(size),w2(1);
    T moyenne,somme=0,ecart_type,Min,Max;
    
    nmax=size;
    size_t=K+1;

    TVecteur<T> t(size_t);
    for (i=0;i<size_t;i++)
	t[i]=i*(size)/K;
// on initialise le vecteur W: W[i]=X[0]+...+X[i]
    for (i=0;i<size;i++)
	{
	  somme+=X[i];  
	  w[i]=somme;
	}
// boucle principale
    for (n0=0;n0<m;n0++)
	for (n=dep[n0];n<nmax;n*=2)
	    {
		somme=long(log10(n));
		w2[0]=0;
		w2.resize(n+1);
		size_t=K+((1-n)*K)/size;
		for (i=0;i<size_t;i++)
		    {
			j=long(t[i])+1;
			k=long(t[i])+n;
			moyenne=long(X.moyenne(j,k));
			for (j=1;j<=n;j++)
			    w2[j]=w[t[i]+j]-w[t[i]]-j*moyenne;
			
			ecart_type=long(sqrt(X.variance(t[i]+1,t[i]+n)));
			Min=w2.min();
			Max=w2.max();
			if (ecart_type!=0)
			    output << somme <<" "<< log10((Max-Min)/ecart_type) <<"\n";
		    }
	    }
    return 0;
}



template<class T> TVecteur<T> TVecteur<T>::inter2arrive(TVecteur<T> x)
{
  TVecteur<T> y(x.maxsize()+1);

  y[0]=0;
  for ( long i=0;i<x.maxsize();i++)
    {
      y[i+1]=y[i]+x[i];
    }
  return y;
}



/*  
//istream& operator>>(istream& s, TVecteur& v)
{
  long n=0,d=1;
  char c=0;

  s >> n >> c;
  if (c=='/')
    {
      s >> d;
    };
  if (s) v=Rationnel(n,d);
  return s;
};
*/




