/***********************************************************/
/* WCPsimul: a Web Cache removal Policy simulation tool    */
/* Copyright (C) 1997 Nicolas Niclausse                    */
/*                                                         */
/* This program is free software; you can redistribute     */
/* it and/or modify it under the terms of the GNU General  */
/* Public License as published by the Free Software        */
/* Foundation; either version 2 of the License, or (at     */
/* your option) any later version.                         */
/*                                                         */
/* This program is distributed in the hope that it will    */
/* be useful, but WITHOUT ANY WARRANTY; without even the   */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A    */
/* PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details.                                       */
/*                                                         */
/* You should have received a copy of the GNU General      */
/* Public License along with this program; if not, write   */
/* to the Free Software Foundation, Inc., 675 Mass Ave,    */
/* Cambridge, MA 02139, USA.                               */
/*                                                         */
/* For information, you can contact me at:                 */
/* Nicolas.Niclausse@sophia.inria.fr                       */
/***********************************************************/
// Classe de Vecteurs ; Descendant de la classe TTableau<T>
//
// Nicolas Niclausse
// Date de cration: 16/10/95

#include "TTableau.h"
#include <iostream.h>

#ifndef TVECTEUR_H
#define TVECTEUR_H

template<class T> class TVecteur: public TTableau<T>
{	
public:	

  TVecteur():TTableau<T>(){};          // Constructeur par defaut
  TVecteur( long n):TTableau<T>(n){};    // Constructeur
  TVecteur(const TVecteur<T>& m):TTableau<T>(m){}; //Constructeur -> arguments de fonctions

  T somme() const;
  T somme( long j) const; // somme des j premiers termes
  T max() const;
  T min() const;
  double moyenne() const;
  double moyenne(const  long k,const  long j) const;  //moyenne de k a j
  double variance() const;
  double variance( long i, long j) const;
  double ecart_type() const;


  TVecteur<T> tri_fusion();

  TVecteur<T> operator + (const TVecteur<T>& m);
  TVecteur<T> operator - (const TVecteur<T>& m); 

  TVecteur<T> operator / (const T& m) const ;
  TVecteur<T> operator * (const T& m) const ;

  int graphique_variance(ostream& output, long,  long);
  int RS(ostream& output);

  TVecteur<T> inter2arrive(TVecteur<T> x);
  TVecteur<T> nombre_clients_GD1(T c, T init);
  TVecteur<T> departs_GD1( T c,T init);
  TVecteur<T> nombre_clients_GD1B(T B, T c, T init);
  TVecteur<T> departs_GD1B(T B, T c, T init);
  double lossrate_GD1B(T B,T c, T init);
  TVecteur<T> temps_sejour(const TVecteur<T>& arrivee,const TVecteur<T>& service, T c);
  TVecteur<T> departs_GG1(const TVecteur<T>& arrivee,const TVecteur<T>& service, T c);
  TVecteur<T> box2ps(const TVecteur<T>& x);
  TVecteur<T> ps2box(const TVecteur<T>& x);

  void initialise_sum(const T& m); // l'element i contient i+T 

};
#endif


